import six
from ldap.dn import str2dn

from .membership import RolesNode as g
from .visitors import PermissionChecker


def group_names(group_dn_list):
    return {str2dn(group)[0][0][1].lower() for group in group_dn_list}


def user_in_group(user, group, **kwargs):
    """Check if a user is in a desired group.

    Args:
        user: A user with its ldap_user property populated (which means the
              user was logged in)
        group: The expected group, as a string or a RolesNode. If you want
               to check membership in several groups, this must be a RolesNode.
        kwargs: passed to the PermissionChecker.visit method.
    """
    user_groups = set()
    if hasattr(user, 'ldap_user'):
        user_groups = group_names(user.ldap_user.group_dns)
    if isinstance(group, six.string_types):
        group = g(group)
    return PermissionChecker(user_groups).visit(group, **kwargs)


def _get_gate(fn):
    from .permissions import requires
    if isinstance(fn, requires):
        return fn.gate
    else:
        return getattr(fn, '_gate', None)


def has_permission(fn, user, permission):
    """Check if the given user has permission to access the given fn.

    This recurses on all nested Gates, unlike the Gate.user_has_permission
    methods which only operate on a single Gate.

    Args:
        fn: The function which may be protected by baya
        user: The django User which is being checked for access
        permission: The permission to check for. One of 'get', 'post', 'any'
    Returns True if the user has permission, False otherwise.
    """
    gate = _get_gate(fn)
    if gate:
        if permission == 'get':
            check_perm = gate.user_has_get_permission
        elif permission == 'post':
            check_perm = gate.user_has_post_permission
        elif permission == 'any':
            check_perm = gate.user_has_any_permission
        else:
            raise ValueError(
                "%s is not a valid permission to check." % permission)
        if check_perm(user):
            if not hasattr(fn, '__closure__'):
                # Not wrapping any function, so bail out
                return True

            nested_fn = fn.__closure__[0].cell_contents
            if not _get_gate(nested_fn):
                return True
            else:
                return has_permission(nested_fn, user, permission)
    return False



"""Helpers for mock_ldap.

You can use these helpers to create a mock LDAP directory for local testing.
See the testsite.directory module for an example.
"""
from collections import defaultdict

DEFAULT_DC = "dc=test"


def person_dn(name, dc=None):
    dc = dc or DEFAULT_DC
    return ("cn=%s,ou=People,%s" % (name, dc)).lower()


def group_dn(name, dc=None):
    dc = dc or DEFAULT_DC
    return ("cn=%s,ou=Access,%s" % (name, dc)).lower()


def person(name, memberOf=None, dc=None, password=None):
    if memberOf is None:
        memberOf = []
    dc = dc or DEFAULT_DC
    domain_parts = dc.split(",")
    email_suffix = ".".join(part.replace("dc=", "") for part in domain_parts)
    password = password or 'password'
    return (person_dn(name, dc=dc), {
        "uid": [name],
        "cn": [name],
        "objectClass": [
            "person", "organizationalPerson", "inetOrgPerson", "posixAccount"],
        "userPassword": [password],
        "uidNumber": ["1000"],
        "gidNumber": ["1000"],
        "givenName": [name.title()],
        "sn": [name],
        "memberOf": memberOf,
        "mail": ["%s@%s" % (name.replace(" ", '-'), email_suffix)],
    })


def group(name, member, memberOf=None, dc=None):
    if memberOf is None:
        memberOf = []
    return (group_dn(name, dc=dc), {
        "cn": [name],
        "objectClass": ["groupOfNames"],
        "member": member,
        "memberOf": memberOf,
    })


def mock_ldap_directory(
        ldap_dc="dc=example,dc=com",
        bind_user="auth",
        bind_password=None,
        extra_users=None,
        group_lineage=None,
        default_password=None):
    """Get a directory for use with MockLdap.

    Args:
        ldap_dc: The base dc to use, eg 'dc=example,dc=com'
        bind_user: The mock user to use for ldap BIND
        bind_password: The password to use for the bind user. Defaults to
            'password'.
        extra_users: A list of (user, group) pairs that will be added to the
            mockldap config.
        group_lineage: A list of (group_parent, group_child) tuples to
            establish a group hierarchy.
        default_password: Set the password for every user to
            `default_password`. Defaults to 'password'.
    Returns a dict.
    """
    extra_users = extra_users or []
    group_lineage = group_lineage or []

    top = (ldap_dc, {"dc": ldap_dc})
    people = ("ou=People,%s" % ldap_dc, {"ou": "People"})
    access = ("ou=Access,%s" % ldap_dc, {"ou": "Access"})
    service_accounts = ("ou=Service-Accounts,%s" % ldap_dc,
                        {"ou": "Service-Accounts"})
    auth_user = person(bind_user, [], ldap_dc, bind_password)

    all_users = []
    all_groups = []

    users_to_groups = defaultdict(list)
    groups_to_users = defaultdict(list)
    # Create (user -> [group]) and (group -> [user]) mappings
    for user, group_name in extra_users:
        users_to_groups[user].append(group_name)
        groups_to_users[group_name].append(user)

    child_group_lineage_dict = defaultdict(list)
    parent_group_lineage_dict = defaultdict(list)
    # Create parent group <-> child group mappings
    for parent, child in group_lineage:
        child_group_lineage_dict[parent].append(child)
        parent_group_lineage_dict[child].append(parent)

    # Create a person for every user, which includes that user's groups
    for user, groups in users_to_groups.items():
        all_users.append(person(user, groups, ldap_dc, default_password))

    # Now we need to build groups for every group we know about
    # First look at all of the groups which have users
    for group_name, users in groups_to_users.items():
        people_in_group = [person_dn(user, ldap_dc) for user in users]
        parent_groups = [
            group_dn(parent_group, ldap_dc)
            for parent_group in parent_group_lineage_dict[group_name]]

        all_groups.append(
            group(group_name,
                  people_in_group + parent_groups,
                  [group_dn(child_group, ldap_dc)
                   for child_group in child_group_lineage_dict[group_name]],
                  dc=ldap_dc))
    # Second, look at the groups that are only in a group hierarchy
    for child, parents in parent_group_lineage_dict.items():
        if child in groups_to_users:
            # Skip it if we've already covered it above
            continue
        all_groups.append(
            group(child,
                  [group_dn(parent, ldap_dc) for parent in parents],
                  [group_dn(child_group, ldap_dc)
                   for child_group in child_group_lineage_dict[child]],
                  dc=ldap_dc))

    directory = [top, people, access, service_accounts, auth_user]
    directory.extend(all_users)
    directory.extend(all_groups)
    return dict(directory)


def mock_ldap_setup(
        ldap_dc="dc=example,dc=com", bind_user="auth",
        **kwargs):
    """Mock ldap setup.

    Args:
        ldap_dc: The base dc to use, eg 'dc=example,dc=com'
        bind_user: The mock user to use for ldap BIND

    Returns an instance of MockLdap. To apply the settings you must call
    `.start()` on the returned value.
    """
    from django.conf import settings
    import mock
    import mockldap

    class MockLDAP(mockldap.MockLdap):
        def start(self, *args, **kwargs):
            super(MockLDAP, self).start(*args, **kwargs)
            self._reconnect_patcher = mock.patch(
                'ldap.ldapobject.ReconnectLDAPObject',
                new_callable=lambda: self.initialize)
            self._reconnect_patcher.start()

        def stop(self, *args, **kwargs):
            self._reconnect_patcher.stop()
            super(MockLDAP, self).stop(*args, **kwargs)

    directory = mock_ldap_directory(ldap_dc, bind_user, **kwargs)

    bind_user_dn = person_dn(bind_user, ldap_dc)
    settings.AUTH_LDAP_BIND_DN = bind_user_dn
    settings.AUTH_LDAP_BIND_PASSWORD = \
        directory[bind_user_dn]['userPassword'][0]
    return MockLDAP(directory)



from django.contrib.admin import ModelAdmin

from baya.utils import has_permission
from baya.permissions import ALLOW_ALL
from baya.permissions import requires


class BayaInlineMixin(object):
    """Mixin for using with Django admin.InlineModelAdmin.

    Inlines, unlike everything else protected by baya, don't get special
    read-only permissions because Django.

    The implementation of change views for inlines is baked deep into the
    Django ModelAdmin.change_view. In order to give read only access to
    inlines I would have to reimplement the entire ModelAdmin.change_view.

    BayaInline classes take required groups for CREATE, UPDATE, and DELETE
    operations. See BayaModelAdmin for documentation on those class-level
    constants.
    """
    CREATE = ALLOW_ALL
    UPDATE = ALLOW_ALL
    DELETE = ALLOW_ALL

    def has_add_permission(self, request):
        return has_permission(requires(self.CREATE), request.user, 'post')

    def has_change_permission(self, request, obj=None):
        return has_permission(requires(self.UPDATE), request.user, 'post')

    def has_delete_permission(self, request, obj=None):
        return has_permission(requires(self.DELETE), request.user, 'post')


class BayaModelAdmin(ModelAdmin):
    """
    All of the create, read, update, and delete permissions will be
    AND-ed with any gates already on the ModelAdmin class (eg when the
    class is decorated by requires).

    This class has four class attributes that determine the CRUD permissions:
        CREATE: Permissions necessary to gain the app.add_model django
            permission. You must still have POST permissions to create
            the object.
        READ: Permissions necessary to gain the app.change_model
            django permission. Note that this does not actually grant
            you permission to change the object, but only to view the
            changelist. You must still have GET permissions to view the
            changelist and POST permissions to change the objects.
        UPDATE: Permissions necessary to gain the app.change_model
            django permission and save changes to the model. You must
            still have POST permissions to change the objects.
        DELETE: Permissions necessary to gain the app.delete_model django
            permission. You must still have POST permissions to delete
            the object.
    """
    CREATE = ALLOW_ALL
    READ = ALLOW_ALL
    UPDATE = ALLOW_ALL
    DELETE = ALLOW_ALL

    def __init__(self, *args, **kwargs):
        super(BayaModelAdmin, self).__init__(*args, **kwargs)
        if hasattr(self, '_gate'):
            self._gate.get_requires &= self.READ
        else:
            requires(self.READ).decorate_admin(self)

        self.add_view = requires(self.CREATE)(self.add_view)
        self.changelist_view = requires(get=self.READ, post=self.UPDATE)(
            self.changelist_view)
        self.change_view = requires(get=self.READ, post=self.UPDATE)(
            self.change_view)
        self.delete_view = requires(self.DELETE)(self.delete_view)

    def user_has_add_permission(self, user):
        return has_permission(self.add_view, user, 'post')

    def has_add_permission(self, request):
        return self.user_has_add_permission(request.user)

    def user_has_change_permission(self, user):
        return (has_permission(self.changelist_view, user, 'any') or
                has_permission(self.change_view, user, 'any'))

    def has_change_permission(self, request, obj=None):
        return self.user_has_change_permission(request.user)

    def user_has_delete_permission(self, request):
        return has_permission(self.delete_view, request, 'post')

    def has_delete_permission(self, request, obj=None):
        return self.user_has_delete_permission(request.user)



"""Arbitrary expressions evaluated as an AST."""
from operator import and_
from operator import not_
from operator import or_
from operator import xor


# Abstract base classes for nodes.

class BaseNode(object):
    """Abstract base class for permission nodes."""

    # Unary operators.
    def __invert__(self):
        return InvertNode(self)

    # Binary operators.
    def __and__(self, other):
        return AndNode(self, other)

    def __rand__(self, other):
        return self & other

    def __or__(self, other):
        return OrNode(self, other)

    def __ror__(self, other):
        return self | other

    def __xor__(self, other):
        return XorNode(self, other)

    def __rxor__(self, other):
        return self ^ other

    # Comparisons.
    def __eq__(self, other):
        raise NotImplementedError()

    def __ne__(self, other):
        return not (self == other)

    def __bool__(self):
        raise TypeError(
            "You probably tried to do 'group1 or/and group2' which, due to "
            "python's semantics, only returns one group. Use a bitwise "
            "operator like &, |, or ^ instead.")

    __nonzero__ = __bool__


class ValueNode(BaseNode):
    """A Node which always returns a given value."""
    def __init__(self, value):
        self.value = value

    def __hash__(self):
        return hash((type(self), self.value))

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "ValueNode(%s)" % repr(self.value)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.value == other.value


class RolesNode(BaseNode):
    def __init__(self, *roles):
        self._roles_set = frozenset(role.lower() for role in roles)

    def __hash__(self):
        return hash((type(self), self._roles_set))

    def get_roles_set(self, **kwargs):
        return self._roles_set

    def __and__(self, other):
        if type(other) is type(self):
            # We can reduce the node complexity by simply combining roles sets.
            return self.__class__(*(self._roles_set | other._roles_set))
        return super(RolesNode, self).__and__(other)

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self._roles_set == other._roles_set)

    def __str__(self):
        return '{%s}' % (', '.join(sorted(self._roles_set)))

    def __repr__(self):
        return ('RolesNode(%s)' %
                (', '.join('"%s"' % role for role in sorted(self._roles_set))))


class DynamicRolesNode(RolesNode):
    """DynamicRolesNodes take callables that return a set of groups.

    The primary use case is when you don't know the required LDAP groups
    until runtime. For example, say you have a BlogPost model that is put
    into one of the categories {'Programming', 'Food', 'Travel'}. Each of
    your users is in only one of those categories, so you want to allow
    posting and editing of content only if the user is authorized for that
    category. You can do this with a DynamicRolesNode:

        class Category(models.Model):
            name = models.CharField(...)

        class BlogPost(models.Model):
            title = models.CharField(...)
            body = models.TextField(...)
            category = models.ForeignKey(Category)

        def _category_name(category_id):
            return {Category.objects.get(id=category_id).name}

        def _blog_post_category(request):
            "Return the category the User is trying to post into."
            # Note that this callable must return a set!
            return _category_name(request.POST['category_id'])

        @requires(post=DynamicRolesNode(_blog_post_category))
        def create_post(request):
            # The user's group membership will be checked on POST to verify
            # that they are in the <category> LDAP group.
            ...

        def _edit_post_category(request):
            return _category_name(BlogPost.objects.get(
                id=request.GET['post_id']).category_id)

        @requires(DynamicRolesNode(_edit_post_category))
        def edit_post(request, post_id):
            # The user can only edit posts in their authorized category.
            ...

    The callable should take kwargs. Note that for Django requests, the
    kwarg `request` will be populated with the current request. The callable
    should return a set of group names.

    Note that DynamicRolesNodes are not compatible with Django admin panels
    because the request is not available to the auth backend when checking
    permissions.
    """
    def __init__(self, *roles_callables):
        self._roles_set = frozenset(roles_callables)

    def get_roles_set(self, **kwargs):
        roles = set()
        for _callable in self._roles_set:
            result = _callable(**kwargs)
            if not isinstance(result, set):
                raise RuntimeError("The callable must return a set.")
            roles |= result
        return roles

    def __str__(self):
        return '{%s}' % self._roles_set

    def __repr__(self):
        return 'DynamicRolesNode(%s)' % self._roles_set


# Operator nodes.

class OperatorNode(BaseNode):
    # Children must define a string that represents this operator.
    display_name = None
    # Children must define an integer arity attribute.
    arity = None
    # Children must define an operator callable with the same arity as
    # specified above.
    operator = None

    def __init__(self, *operands):
        if len(operands) != self.arity:
            raise ValueError('Incorrect number of operands for %s: %d.  '
                             'Expected %d.' % (self.__class__.__name__,
                                               len(operands), self.arity))
        for operand in operands:
            if not isinstance(operand, BaseNode):
                raise TypeError('%r is not a child of BaseNode.' %
                                operand)
            # TODO: Check if operand is an instance of basestring and cast to
            # a BaseNode?
        self._operands = tuple(operands)

    def __hash__(self):
        return hash((type(self), self._operands))

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self._operands == other._operands)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ', '.join(map(repr, self._operands)))


class InvertNode(OperatorNode):
    display_name = '~'
    arity = 1
    operator = not_


class AndNode(OperatorNode):
    display_name = '&'
    arity = 2
    operator = and_


class OrNode(OperatorNode):
    display_name = '|'
    arity = 2
    operator = or_


class XorNode(OperatorNode):
    display_name = '^'
    arity = 2
    operator = xor



from functools import partial
import six

from django.conf import settings
from django.contrib.auth import get_permission_codename
from django_auth_ldap.backend import LDAPBackend


class NestedLDAPGroupsBackend(LDAPBackend):

    use_reconnecting_client = getattr(
        settings, 'BAYA_USE_RECONNECTING_CLIENT', False)

    def _get_ldap(self):
        if self.use_reconnecting_client:
            if self._ldap is None:
                ldap_module = super(NestedLDAPGroupsBackend, self).ldap
                self._ldap = ReconnectingLDAP(ldap_module)
            return self._ldap
        else:
            return super(NestedLDAPGroupsBackend, self).ldap
    ldap = property(_get_ldap)

    def get_all_permissions(self, user, obj=None):
        """Return a set of <app_label>.<permission> for this user.

        Note that these results will only be applicable to the admin panel.

        <permission> is something like add_mymodel or change_mymodel.
        """
        if hasattr(user, '_baya_cached_all_permissions'):
            return user._baya_cached_all_permissions
        permissions = super(NestedLDAPGroupsBackend, self).get_all_permissions(
            user, obj)
        from baya.admin.sites import _admin_registry
        for admin_site in _admin_registry:
            for model, opts in six.iteritems(admin_site._registry):
                app = model._meta.app_label
                perm_name = partial(get_permission_codename, opts=model._meta)
                if (hasattr(opts, 'user_has_add_permission') and
                        opts.user_has_add_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('add')))
                if (hasattr(opts, 'user_has_change_permission') and
                        opts.user_has_change_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('change')))
                if (hasattr(opts, 'user_has_delete_permission') and
                        opts.user_has_delete_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('delete')))
        user._baya_cached_all_permissions = permissions
        return permissions


class ReconnectingLDAP(object):
    """
    An object that makes the standard ldap.initialize function use
    ldap.ldapobject.ReconnectLDAPObject instead.

    The django_auth_ldap library does not provide a simple way to
    override how the LDAP connection is created. The actual method
    that initiates the connection is on the _get_connection method
    on the hidden _LDAPUser class.

    Instances of this class take the original ldap module instance
    and provide the same set of attributes, except that the initialize
    attribute returns ReconnectLDAPObject instead of the original
    initialize function (which just instantiates a SimpleLDAPObject).
    """
    retry_max = 6

    def __init__(self, original_module):
        self._original_module = original_module

    def __getattr__(self, name):
        """
        Return the requested value from the original module.

        This is called when ordinary attribute lookups fail.
        Since initialize is defined below, it will not be called
        for accesses to that attribute.
        """
        return getattr(self._original_module, name)

    def initialize(self, uri):
        return self._original_module.ldapobject.ReconnectLDAPObject(
            uri, retry_max=self.retry_max)



import os

from pip.download import PipSession
from pip.req import parse_requirements
from setuptools import find_packages
from setuptools import setup


def load_readme():
    PROJECT_DIR = os.path.dirname(__file__)
    readme_file = "README.md"
    try:
        return open(os.path.join(PROJECT_DIR, readme_file), 'r').read()
    except Exception:
        raise RuntimeError("Cannot find readme file {fname}.".format(
            fname=readme_file))


def load_version():
    """Open and parse out the version number from the _version.py module.

    Inspired by http://stackoverflow.com/a/7071358
    """
    import re
    version_file = "baya/_version.py"
    version_line = open(version_file).read().rstrip()
    vre = re.compile(r'__version__ = "([^"]+)"')
    matches = vre.findall(version_line)
    if matches and len(matches) > 0:
        return matches[0]
    else:
        raise RuntimeError(
            "Cannot find version string in {version_file}.".format(
                version_file=version_file))

version = load_version()
long_description = load_readme()

# Build description from README and build metadata from Go pipeline.
long_description += "\n"
long_description += "build_revision: {}\n".format(os.getenv('GO_REVISION'))
long_description += "build_pipeline: {}\n".format(os.getenv('GO_PIPELINE_NAME'))
long_description += "build_label:    {}\n".format(os.getenv('GO_PIPELINE_LABEL'))

requirements = ['Django'] + [
    str(ir.req)
    for ir in parse_requirements('./requirements.txt', session=PipSession())]

test_requirements = [
    str(ir.req)
    for ir in parse_requirements('./requirements-dev.txt', session=PipSession())]

setup(
    name='baya',
    version=version,
    description="Nested LDAP Groups authorization.",
    long_description=long_description,
    author='Steven Buss',
    author_email='steven.buss@gmail.com',
    maintainer='Counsyl Platform Team',
    maintainer_email='platform@counsyl.com',
    license='MIT',
    url='https://github.com/counsyl/baya',
    download_url=(
        'https://github.com/counsyl/baya/tarball/v%s' %
        version),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
    packages=find_packages(exclude=[
        'tests', 'tests.*']),
    include_package_data=True,
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require={
        'development': [
            'mockldap>=0.2.0',
        ],
    },
)



import logging

logger = logging.getLogger('baya')


class DynamicRoleCallable(object):
    def __call__(self, **kwargs):
        raise NotImplementedError("You must implement this method.")


class DjangoRequestCallable(DynamicRoleCallable):
    def __call__(self, request, **kwargs):
        raise NotImplementedError("You must implement this method.")


class DjangoRequestGroupFormatter(DjangoRequestCallable):
    """Create a dynamic group, whose name depends on a query parameter.

    For example, say you have a view:

        def admin_view(request, thing):
            ...

    If you want to restrict access to this view based on a request.GET query
    parameter or on a named URL pattern kwarg you can use this class. Say you
    want to restrict access to thing='abcd' to only those users in the
    abcd_admin group. Then you can annotate the view like this:

        from baya.membership import DynamicRolesNode
        from baya.dynamic_roles import DjangoRequestGroupFormatter
        @requires(DynamicRolesNode(DjangoRequestGroupFormatter(
            "%s_admin", 'thing')))
        def my_view(request, thing):
            ...

    Note that if you have a url like

        url(r'my_view/(?P<param>\w+)/',
            requires(DynamicRolesNode(DjangoRequestGroupFormatter(
                     "%s_admin", 'thing')))(path.to.my_view),
        )

    Then the named reverse kwarg named `param` will always take precedence over
    a url query parameter named `param`. For example:

        http://mysite.com/my_view/my_param1?param=my_param2

    This url will pass my_param1 to the formatter and log a warning about
    my_param2 being there.
    """
    def __init__(self, group_name_format, query_field):
        """
        Args:
            group_name_format: The group name which can be formatted with
                the query_field's value. EG "%s_admin"
            query_field: The query parameter which contains the value to
                format `group_name_format` with.
        """
        self.group_name_format = group_name_format
        self.query_field = query_field

    def __call__(self, request):
        url_kwargs = request.resolver_match.kwargs
        if self.query_field in url_kwargs and self.query_field in request.GET:
            logger.warning(
                "The requested parameter (%s) was in both the url kwargs "
                "and the url query parameters." % self.query_field)
        value = url_kwargs.get(self.query_field,
                               request.GET.get(self.query_field))
        group = self.group_name_format % value
        return {group.lower()}

    def __repr__(self):
        return "%s(%s, %s)" % (
            self.__class__.__name__,
            self.group_name_format,
            self.query_field)
